
/* ================================================================= */
/*			  TBOOT.C                                    */
/* ================================================================= */

#include <stdlib.h>
#include <stdio.h>
#include <alloc.h>
#include <dos.h>
#include <dir.h>
#include <io.h>
#include "..\..\dtc2.h"

#define BOOT_BUFSIZE	32000u

#define OUTP	outportb
#define INP	inportb

/* Terminate codes */

#define T_TERMINATED       0
#define T_BAD_BOOT         9
#define T_BAD_COMMAND      15
#define T_BAD_COMMAND_LINE 16
#define T_BAD_RECORD       17
#define T_BAD_INT32        18
#define T_BAD_HOST_INIT    19
#define T_BAD_PROTOCOL	   20

/* Boot error codes */

#define BOOT_FILE_OK           0
#define BOOT_FILE_OPEN_ERR     1
#define BOOT_FILE_READ_ERR     2
#define BOOT_FILE_SEND_ERR     3
#define BOOT_FILE_CLOSE_ERR    4
#define BOOT_FILE_PROTOCOL_ERR 5
#define BOOT_FILE_TARGET_ERR   6
#define BOOT_FILE_VERSION_ERR  7
#define BOOT_FILE_SCALER_ERR   8
#define BOOT_FILE_VECTOR_ERR   9
#define BOOT_FILE_CODE_ERR    10
#define DEBUG_FILE_OPEN_ERR   11
#define DEBUG_FILE_CLOSE_ERR  12
#define BOOT_BUFFER_ERR       13

typedef enum { sStart,sEnd,sData,sBlock,sSymb,sComm,sNorec } rectyp;
static int Quiet=0;
static int Link=0x150;
static size_t Bufsize;
static int result = BOOT_FILE_OK;
unsigned long *DSPBUF;

int booting;

static char SkipCode[] = {
/* bytes out,messages in: */
/*	db	Endboot-Boot
Boot:	ajw 4		; let space for IN
	stl 0; stl 0; stl 0
	mint; sthf; mint; stlf
	mint; adc $f0000; stl 1
	ldl 1; ldl 0; ldc Endskip-Skipper; in;
	ldl 0		; Creg=link in
	ldl 1		; Breg=Iptr -> 800f0000
	ldl 1
	adc $100	; Areg=Wptr -> 800f0100
	gajw
	rev		; Areg=Iptr Breg=Wptr
	gcall
Endboot:
*/
  0x23,0xb4,0xd0,0xd0,0xd0,0x24,0xf2,0x21,0xf8,0x24,0xf2,0x21,0xfc,0x24,
  0xf2,0x2f,0x20,0x20,0x20,0x80,0xd1,0x71,0x70,0x27,0x46,0xf7,0x70,0x71,
  0x71,0x21,0x20,0x80,0x23,0xfc,0xf0,0xf6,
/*
Skipper:
	or			; kill Areg,Breg
	rev
	mint
	stnl	28		; bootlink(in) = 80000070
	testerr
	clrhalterr
	ldc	0
	stl	1
	ldlp	1
	mint
	ldnl	28
	ldc	1
	in			; input skip link
	ldl	1
	mint
	wsub
	mint
	stnl	29		; newlink(out) = 80000074
	ldc	0
	stl	1
	ldc	11
	stl	2
Btlp1:	mint
	ldl	1
	mint
	wsub
	stnl	0
	ldlp	1
	ldc	Ebtlp1 - Btlp1
	lend
Ebtlp1:	ldc	0
	sttimer
	ldc	SkipIn-Skpl1
	ldpi
Skpl1:	stl	-1
	ldlp	0
	ldpri
	or
	ajw	256
	runp
*/
					  0x24,0xfb,0xf0,0x24,0xf2,0x21,
  0xec,0x22,0xf9,0x25,0xf7,0x40,0xd1,0x11,0x24,0xf2,0x21,0x3c,0x41,0xf7,
  0x71,0x24,0xf2,0xfa,0x24,0xf2,0x21,0xed,0x40,0xd1,0x4b,0xd2,0x24,0xf2,
  0x71,0x24,0xf2,0xfa,0xe0,0x11,0x4b,0x22,0xf1,0x40,0x25,0xf4,0x21,0x4c,
  0x21,0xfb,0x60,0xdf,0x10,0x21,0xfe,0x24,0xfb,0x21,0x20,0xb0,0x23,0xf9,
/*
SkipOut:
  ldlp 1; mint; ldnl 28; ldc 1; in; ldlp 1; mint; ldnl 29; ldc 1; out;
	j	SkipOut
*/
  0x11,0x24,0xf2,0x21,0x3c,0x41,0xf7,0x11,0x24,0xf2,0x21,0x3d,0x41,0xfb,
  0x60,0x00,
/*
SkipIn:
	ldlp	1
	mint
	ldnl	29
	ldnlp	4
	ldc	8	; W[1]=pri .. W[3]=size (buflen>=36)
	in
	ldc	36
	ldl	2	; buf->size
	gt		; A=1 if B>A (36>size), 0 if 36<=size
	cj	Skpi2
	ldc	36
	stl	2
Skpi2:	ldlp	3
	mint
	ldnl	29
	ldnlp	4
	ldl	2
	adc	-8
	in		; rest of message
	ldlp	1
	mint
	ldnl	28
	ldnlp	-4
	ldl	2
	out
	j	SkipIn
Endskip:
*/
	    0x12,0x24,0xf2,0x21,0x3d,0x54,0x48,0xf7,0x22,0x44,0x73,0x25,
  0xfa,0xd1,0xf9,0xa3,0x22,0x44,0xd1,0x14,0x24,0xf2,0x21,0x3d,0x54,0x71,
  0x60,0x88,0xf7,0x12,0x24,0xf2,0x21,0x3c,0x60,0x5c,0x71,0xfb,0x62,0x08
 };


/* ================================================================== */


int DTC2present(void)
{
 int err,HWxmode;

  err=1;
/* reset XMODE */
  outportb(oXMODE,0);
  HWxmode = inportb(iXMODE)&1;	/* 1 if inv(PC), 0 if PROM */

/* toggle/test XMODE */
  outportb(oXMODE,1);
  if ((inportb(iXMODE)&1)^HWxmode) err=0;
  outportb(oXMODE,0);
  if ((inportb(iXMODE)&1)^HWxmode) err=0;
  if (!err) {
/* test echo of CTRL0 bit */
    outportb(FPGA,1);
    if ((inportb(FPGA+1)&0x20)==0) err=1;
    outportb(FPGA,0);
    if ((inportb(FPGA+1)&0x20)!=0) err=1;
   }
  return(!err);
}


/* ----------------------------------------------------------------- */


void resetall(void)
{
  outportb(oDRES,0);	/* reset DSP */
  outportb(oDMODE,0);
  outportb(oTANALY,0);
  outportb(oTRESET,0);
  outportb(oTRESET,1);	/* reset T425 */
  delay(20);
  outportb(oTRESET,0);
  delay(20);
  outportb(FPGA,0);
  outportb(oDRES,1);
}


/* ----------------------------------------------------------------- */
/*	LOD file read						     */
/* ----------------------------------------------------------------- */

static rectyp findkeyw (char *t)
{
 char typessym[6][12] =
      { "START","END","DATA","BLOCKDATA","SYMBOL","COMMENT" };
 int i,found;

  found=0;
  for (i=0; (i<6)&&(!found); i++) {
    found = !stricmp(t,typessym[i]);
   }
  if (!found) return(sNorec);
   else return((rectyp)(i-1));
}


int load56(char *fname)
{
 char t[81];
 int i;
 rectyp blktype = sNorec;
 char mspace[3];
 unsigned int len;
 unsigned long *buf=DSPBUF;
 long start,V;
 FILE *fp;

  if ((fp = fopen(fname,"rt")) == NULL) return(1);
  printf("Loading DSP file %s...",fname);
  memset(DSPBUF,0xff,DBUFSIZ*sizeof(long));

  fprintf(stderr,"\n");
  do {
   if (fscanf(fp,"%s",t)==1) {
     if (*t == '_') {
       blktype = findkeyw(&(t[1]));
       switch (blktype) {
	 case sData:
	     fscanf(fp,"%s",mspace);
	     fscanf(fp,"%4lX",&start);
	     len=0; i=1;
	     buf=DSPBUF+start;
	     while (i) {
	       i = fscanf(fp,"%6lX",&V);
	       if ((toupper(*mspace)=='P') && i) { *buf++=V; len++; }
	      }
	     fprintf(stderr,"\tblk %c:%04lx l=%x\n",*mspace,start,len);
	     break;
	 case sBlock:
	     fscanf(fp,"%s",mspace);
	     fscanf(fp,"%4lX",&start);
	     fscanf(fp,"%4x",&len);
	     fscanf(fp,"%6lX",&V);
	     if (toupper(*mspace)=='P') {
	       buf=DSPBUF+start;
	       fprintf(stderr,"\tfill %06lx %c:%04lx l=%x\n",V,*mspace,start,len);
	       for (i=0; i<len; i++) *buf++=V;
	      }
	     break;
	 default: break;
	}
      }
    }
  } while (!(feof(fp) || (blktype==sEnd)));
 fclose(fp);
 return(0);
}


static int bootDSP(char *fname)
{
 int n,t;
 unsigned long *p;
 unsigned long v;

  if (load56(fname)) return(1);

  p = DSPBUF;
  for (n=0; n<512; n++) {
    v=*p++;
    t=10000; while (((inp(HI_ISR)&0xf2)==0) && (--t));
    if (t==0) { printf("#%d\n",n); return(1); }
    disable();
    outp(HI_TXM,(v>>8)); outp(HI_TXM,(v>>8));
    outp(HI_TXH,(v>>16)); outp(HI_TXH,(v>>16));
    outp(HI_TXL,v);
    enable();
   }
  outportb(OnCEstat,0);		/* set DR=1, clear ACK */
  delay(1);	/* wait for DSP to reinitialize */
  return(0);
}


/* ================================================================== */


static int near boot_message(int *l,char *m)
{
 unsigned long t0=0xfffffl;
 unsigned long t;

  while (booting&&((*l)--)) {
    t=t0; while (((INP(Link+3)&1)==0)&&(--t));
    if (t==0) return(0);
    OUTP(Link+1,*m++);
   }
  return(1);
}


static int near load_code (char *bootf)
{
 static char Flicks[]="|/-\\|/-\\" ;
 int Flick=0;
 FILE *fs;
 char *buf = NULL;
 int  count = 0;
 int  fh;
 long skip;

  if ((fs = fopen(bootf,"rb")) != NULL) {

    fh = fileno(fs);
    if ((buf=(char *)malloc(Bufsize))==NULL)
      result = BOOT_BUFFER_ERR;

    skip=getc(fs);
    if (skip==0xf0) {		/* INMOS std boot */
      lseek(fh,0,SEEK_SET);
     }
    else {			/* TXE file */
      lseek(fh,skip+1,SEEK_SET);
      read(fh,&skip,sizeof(long)); lseek(fh,skip,SEEK_CUR);
     }

    while (result == BOOT_FILE_OK) {
      count = read(fh,buf,Bufsize);
      if (count > 0) {
	if (boot_message(&count,buf)==0)
	  result = BOOT_FILE_SEND_ERR;
	 else if (!Quiet) {
	   fprintf(stderr,"%c\b", Flicks[Flick]);
	   if (++Flick == 8) Flick -= 8;
	  }
       }
      else {
	if (eof(fh) == 0) result = BOOT_FILE_READ_ERR;
	 else break;
       }
     }
    if (result == BOOT_FILE_OK) {
      if (fclose(fs) != 0) result = BOOT_FILE_CLOSE_ERR;
     }
    else
      (void)fclose(fs);
   }
  else
    result = BOOT_FILE_OPEN_ERR;

  if (buf) free(buf);
  return(count);
}


/* ================================================================== */


int net_boot (char *bootf,int link,int iskip,int quiet)
{
 int i,k,l;

  if (bootf != NULL) {

    Bufsize = BOOT_BUFSIZE;
    Quiet = quiet;
    Link = link;
    result=BOOT_FILE_OK;
    booting=1;

    if (!Quiet)
      fprintf(stderr, "Booting transputer network ... ");

    while (iskip) {
      l=sizeof(SkipCode);
      if (boot_message(&l,SkipCode)==0) {
	result = BOOT_FILE_SEND_ERR; k=sizeof(SkipCode)-l;
	break;
       }
      while ((INP(Link+3)&1)==0);
      OUTP(Link+1,((iskip&7)-1));
      iskip>>=4;
     }

    if (result == BOOT_FILE_OK)
      k = load_code(bootf);

    if (!Quiet)
      fprintf(stderr, "\r                                   \r");

    if (result == BOOT_FILE_OK) return(1);
    switch (result) {
      case BOOT_FILE_OPEN_ERR :
	fprintf(stderr, "\nFailed to open boot file : %s\n", bootf);
	break;
      case BOOT_FILE_READ_ERR :
	fprintf(stderr, "\nFailed to read from boot file : %s\n", bootf);
	break;
      case BOOT_FILE_SEND_ERR :
	fprintf(stderr, "\nFailed to transmit boot file : %s, byte %d\n",
		bootf,k);
	break;
      case BOOT_FILE_CLOSE_ERR :
	fprintf(stderr, "\nFailed to close boot file : %s\n", bootf);
	break;
      default:
	fprintf(stderr, "\nUnknown error in booting %s\n", bootf);
	break;
     }
    return(0);
   }
  else
    return(1);
}

/* ================================================================== */


int main(int argc,char *argv[])
{
 char fn[MAXPATH];
 char drive[MAXDRIVE];
 char dir[MAXDIR];
 char file[MAXFILE];
 char ext[MAXEXT];
 int i,flags;
 char far *QR;

  if (!DTC2present()) {
    printf("DTC2 not found\n");
    exit(1);
   }
  if (argc<3) {
    fprintf(stderr,"Use: %s <t-file> <dsp-file>\n",argv[0]);
    exit(1);
   }
  DSPBUF=(ulong *)malloc(DBUFSIZ*sizeof(long));
  QR=MK_FP(0xcc00,0);

  resetall();
  for (i=0; i<512; i++) QR[i]=0;
  QR[0x200]=0; QR[0x201]=0;

  flags=fnsplit(argv[1],drive,dir,file,ext);
  if (!(flags&EXTENSION)) strcpy(ext,".btl");
  fnmerge(fn,drive,dir,file,ext);
  if (!net_boot(fn,0x150,0,0)) {
    fprintf(stderr,"T5 boot failed\n"); exit(1);
   }
  flags=fnsplit(argv[2],drive,dir,file,ext);
  if (!(flags&EXTENSION)) strcpy(ext,".lod");
  fnmerge(fn,drive,dir,file,ext);
  if (bootDSP(fn)) {
    fprintf(stderr,"DSP boot failed\n"); exit(1);
   }
  QR[0x201]=1;
  outportb(0x162,1);
  outportb(FPGA,0xa0);

  return(0);
}


/* ================================================================== */

