
#include "..\dtc2.h"
#include <dir.h>

void dspwrite(unsigned long v);
void dspwrite8(unsigned v);
int dspwrite_t(unsigned long v,long t);
long dspread(void);
long dspread_t(long t);
int DSPview(void);
int bootDSP(char *fname);
int load56(char *fname);

unsigned long *DSPBUF;
char fn[] = "dtc2x.bit";
char dspfn[] = "flash.lod";
char dataf[256];
int mon56,ispause,erase;
int HWxmode;

char far *QRAM;

/* ----------------------------------------------------------------- */


void PAUSE(void)
{
  if (ispause) {
    printf("--Paused--\n");
    getch();
   }
}


int DTC2present(void)
{
 int k,seg,err;
 char *es;

  err=1;
/* reset XMODE */
  outportb(oXMODE,0);
  HWxmode = inportb(iXMODE)&1;	/* 1 if inv(PC), 0 if PROM */

/* toggle/test XMODE */
  outportb(oXMODE,1);
  if ((inportb(iXMODE)&1)^HWxmode) err=0;
  outportb(oXMODE,0);
  if ((inportb(iXMODE)&1)^HWxmode) err=0;

  QRAM=NULL;
  es=getenv("DTCQRAM");
  if (es==NULL) err=1;
  else {
    if (sscanf(es,"%x",&k)==1) {
      switch(k) {
	case 0: case 0xc8: case 0xc800: seg=0xc800; break;
	case 1: case 0xcc: case 0xcc00: seg=0xcc00; break;
	case 2: case 0xd0: case 0xd000: seg=0xd000; break;
	default: err=1;
       }
      if (!err) QRAM=MK_FP(seg,0);
     }
   }
  if (QRAM==NULL)
    printf("Invalid DTCQRAM environment setting\n");

  return(!err);		/* err=0 if present */
}


/* ----------------------------------------------------------------- */


void resetall(void)
{
  outportb(oECF2,1);
  outportb(oECF3,0);
  outportb(oDRES,0);	/* reset DSP  */
  outportb(oDMODE,0);
}


/* ------------------------------------------------------------------- */


int putflash(char *fn,int segm)
{
 int i;
 long v;
 unsigned long *p;

  if (bootDSP(dspfn)) {
    printf("Error loading %s\n",dspfn); return(1);
   }
  if ((v=dspread())!=0x29f010l) {
    printf("Wrong DSP code, returned %06lx\n",v);
    return(1);
   }
  if ((v=dspread())!=0x2001l) {
    printf("Wrong Flash code, returned %06lx\n",v);
    return(1);
   }
  printf("Erasing block %d...",segm); v=segm;
  dspwrite(v);
  v=dspread();
  if (v==(0x434c00l|segm)) printf("done %06lx\n",v);
   else {
    printf("Erase returns %06lx\n",v); return(1);
   }
  if (erase) return(0);

  if (load56(fn)) {
    printf("File %s not loaded\n",fn); return(1);
   }
  printf("Transferring %s...",fn);
  p=DSPBUF;
  i=0;
  while (i<16384) {
    v=*p++;
    dspwrite8(v&0xff); i++; if (i>=16384) break;
    dspwrite8((v>>8)&0xff); i++; if (i>=16384) break;
    dspwrite8((v>>16)&0xff); i++; if (i>=16384) break;
   }
  printf("result=%06lx\n",dspread());

  return(0);
}


/* ------------------------------------------------------------------- */


int main (int argc, char *argv[])
{
 int err,i,j,k,n,go;
 char *p;
 char drive[MAXDRIVE];
 char dir[MAXDIR];
 char file[MAXFILE];
 char ext[MAXEXT];
 int flags,segm;
 long v;

  printf("DTC load Flash ["__DATE__" "__TIME__"]\n");
  mon56=0;
  ispause=0; erase=0; go=0;

  DSPBUF=(unsigned long *)malloc(DBUFSIZ*sizeof(long));
  if (DSPBUF==NULL) exit(-1);

  k=argc; *dataf=0; segm=-1;
  while (--k) {
    p=argv[k];
    if ((*p=='-')||(*p=='/')) {
      switch(p[1]) {
	case 'e': erase=(p[2]!='-'); break;
	case 'g': go=(p[2]!='-'); break;
	case 'p': ispause=(p[2]!='-'); break;
	case 'm': mon56=(p[2]!='-'); break;
	case 's': if (sscanf(p+2,"%d",&j)==1) segm=j;
		  break;
	default:  break;
       }
     }
    else strcpy(dataf,p);
   }
  if ((segm<0)||(segm>7)) {
    printf("Segment must be 0-7, not %d\n",segm);
    *dataf=0;		/* fall thru */
   }
  if ((*dataf==0)&&(!erase)) {
    printf("Use: DTCFLASH <file{.lod}> -s#segm <opt>\n");
    exit(1);
   }

  resetall();

  if (!DTC2present()) {
    printf("DTC2 not found\n");
    exit(1);
   }
  k = inportb(iXDONE)&1;
  if (!k) { printf("FPGA not configured\n"); exit(1); }

  if (!erase) {
    flags=fnsplit(dataf,drive,dir,file,ext);
    fnmerge(dataf,drive,dir,file,".lod");
   }
  putflash(dataf,segm);

  if (go) {
    outportb(oDRES,0);
    outportb(oDMODE,1);
    printf("Restarting DSP from Flash says ");
    outportb(oDRES,1);
    delay(500);
    v=dspread_t(200000); printf("%06lx\n",v);
    if (v!=0x56002l) {
      outportb(oDRES,0);	/* reset DSP  */
      outportb(oDMODE,0);
      exit(1);
     }
    if (mon56) DSPview();
   }
  else {
    outportb(oDRES,0);	/* reset DSP  */
    outportb(oDMODE,0);
   }

  return(0);
}

