
/************************************************************************
*									*
*	Device Driver for DSP56002/DTC2                                 *
*									*
*************************************************************************/

#ifdef MODULE
# include <linux/config.h>
# include <linux/module.h>
#endif
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/types.h>
#include <linux/ioport.h>
#include <linux/major.h>

#include <asm/io.h>
#include <asm/segment.h>

#include <linux/dtc2.h>
#include "OnCE.h"

#define CARD_NAME	"dtc2"

#ifdef DSP_DEBUG
#define DEB(x) x
#define PRINTK(f, a...)	printk(f, ##a)
#define NO_DEB(x) /* */
#else
#define DEB(x) /* */
#define PRINTK(f, a...)
#define NO_DEB(x) x
#endif

#ifdef DSP_DEBUG_MORE
#define DEB_MORE(x) x
#define PRINTKM(f, a...)	printk(f, ##a)
#else
#define DEB_MORE(x) /* */
#define PRINTKM(f, a...)
#endif

static int use_schedule = 0;            /* don't use schedule() in init() */
static int DSPexists = 0;
static int DSPowner = 0;
static int OnCEmode = 0;
/*static int preinit = 0;*/
static int HWxmode = 0;
static int FPGAerr,FPGAdone,FPGAmode;

/* map DSP exchange area into QRAM */
static Qrdx *QRDX = (Qrdx *)QR_PAGE3;

#define TCNT1	20000
#define TCNT2	10000

extern int fpgaconfTAB(void);

static int dsp_flags[16] = { 0, 0 };

/****************************************************************************/


void d_delay( int il )
{
	if ( use_schedule && !il )
	{
		DEB(printk("dsp_delay: Using schedule()\n");)
		current->state = TASK_INTERRUPTIBLE;
		current->timeout = jiffies + 10;   /* 100ms */
		schedule();
	}
	else
	{	/* only for detection in xxx_init() */
		volatile int i;
		DEB(printk("dsp_delay: Using idle loop\n");)
		for (i = 0; i < 100000; i++);
	}
}


/* -------------------------------------------------------------------- */


static int DTC2present(void)
{
  int err;

	err=1;
/* reset XMODE */
	outb(0,oXMODE);
	HWxmode = inb(iXMODE)&1;	/* 1 if inv(PC), 0 if PROM */

/* toggle/test XMODE */
	outb(1,oXMODE);
	if ((inb(iXMODE)&1)^HWxmode) err=0;
	outb(0,oXMODE);
	if ((inb(iXMODE)&1)^HWxmode) err=0;
	return(!err);
}


/****************************************************************************
 *
 */
 
static int read_dmem(struct inode * inode, struct file * file, char * buf, int count)
{
	unsigned short p = file->f_pos & 0x1ff;
	int c1, c2, read;

	if ((count < 0) || (count > 512))
		return -EINVAL;
/*	if (MAP_NR(p) >= MAP_NR(high_memory)) return 0; */
	c1 = min (count, (512 - p));
	c2 = count + p;
	read = 0;
	if (c1) memcpy_tofs(buf, (void *) (QR_PAGE0 + p), c1);
	if (c2 > 512) memcpy_tofs(buf+c1, (void *) QR_PAGE0, (c2 & 0x1ff));
	read += count;
	file->f_pos += read;
	return read;
}


static int write_dmem(struct inode * inode, struct file * file, const char * buf, int count)
{
	unsigned short p = file->f_pos & 0x1ff;
	int c1, c2, written;

	if ((count < 0) || (count > 512))
		return -EINVAL;
/*	if (MAP_NR(p) >= MAP_NR(high_memory)) return 0; */
	c1 = min (count, (512 - p));
	c2 = count + p;
	written = 0;
	if (c1) memcpy_fromfs((void *) (QR_PAGE0 + p), buf, c1);
	if (c2 > 512) memcpy_fromfs((void *) QR_PAGE0, buf+c1, (c2 & 0x1ff));
	written += count;
	file->f_pos += written;
	return count;
}


/****************************************************************************/


static void TXwrite(DSPFix24 val)
{
 unsigned count = TCNT1;

  while ((inb(HI_ISR)&2)==0) {
    if ((--count)==0) { count=TCNT1; schedule(); }
  }
  _TXwrite(val);
}


static DSPFix24 RXread(void)
{
 unsigned count = TCNT1;

  while ((inb(HI_ISR)&1)==0) {
    if ((--count)==0) { count=TCNT1; schedule(); }
  }
  return(_RXread());
}


static int DSPcommand(unsigned char c)
{
 unsigned count=TCNT1,ntest;

  while (inb(HI_CVR)&0x80) {
    if ((--count)==0) { count=TCNT1; schedule(); }
  }
  outb((c|0x80), HI_CVR);
  count = TCNT1; ntest=5;
  while ((inb(HI_CVR)&0x80)&&ntest) {
    if ((--count)==0) { count=TCNT1; ntest--; schedule(); }
  }
  return(ntest==0? -1:0);
}


int DSPreset (void)
{
	QRDX->ident=0xffffff;
	outb(0,oDRES);
	outb(1,oECF2);
	outb(0,oECF3);
	outb(1,oDMODE);
	d_delay(1);
	outb(1,oDRES);
	d_delay(0);
	return(_RXread());
}


/****************************************************************************
 *
 */
static int dsp_read(struct inode * inode, struct file * file,
			char * buf, int count)
{
	const unsigned int minor = MINOR(inode->i_rdev);

	if (minor > 15)
	  return read_dmem (inode, file, buf, count);

	PRINTKM("DSP(%d): reading %d bytes (polling mode).\n", minor, count);

/*	if ( count < 0)
	{
		PRINTK("DSP: invalid argument for reading: count = %d.\n", count); */
		return -EINVAL;
/*	} */
}


/****************************************************************************
 *
 */
static int dsp_write(struct inode * inode, struct file * file,
			char * buf, int count)
{
	const unsigned int minor = MINOR(inode->i_rdev);

	if (minor > 15)
	  return write_dmem (inode, file, buf, count);

	PRINTKM("DSP(%d): writing %d bytes (polling mode).\n", minor, count);

/*	if ( count < 0)
	{
		PRINTK("DSP: invalid argument for reading: count = %d.\n", count); */
		return -EINVAL;
/*	} */
}


/****************************************************************************
 *
 */
static int dsp_lseek(struct inode * inode, struct file * file,
		    off_t offset, int origin)
{
	const unsigned int minor = MINOR(inode->i_rdev);
	
	if (minor < 16)	return -ESPIPE;
	
	switch (origin) {
		case 0:
			file->f_pos = offset;
			return file->f_pos;
		case 1:
			file->f_pos += offset;
			return file->f_pos;
		default:
			return -EINVAL;
	}
	if (file->f_pos < 0)
		return 0;
	return file->f_pos;
}


/****************************************************************************
 *
 */
static int dsp_open(struct inode * inode, struct file * file)
{
	const unsigned int minor = MINOR(inode->i_rdev);

	if (minor > 15) return 0;

	if (DSP_F(minor) & DSP_BUSY)
	{
		PRINTK("DSP(%d) not opened, device busy.\n", minor);
		return -EBUSY;
	}

	DSP_F(minor) |= DSP_BUSY;

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif

	PRINTK( "DSP(%d) opened.\n", minor);
	return 0;

}


/****************************************************************************
 *
 */
static void dsp_release(struct inode * inode, struct file * file)
{
	const unsigned int minor = MINOR(inode->i_rdev);

	if (minor > 15) return;

	DSP_F(minor) &= ~DSP_BUSY;

#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif

	PRINTK("DSP(%d) released.\n", minor);

}


/****************************************************************************
 *
 */
static int dsp_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg)
{
	const unsigned int minor = MINOR(inode->i_rdev);
/*	int result = arg; */

	PRINTK("DSP(%d) ioctl, cmd: 0x%08x, arg: 0x%08x.\n", minor, cmd, (int) arg);

/*	if (minor > 15) */ return -EINVAL;
	
/*	return result; */
}


/****************************************************************************/

static struct file_operations dsp_fops = {
	dsp_lseek,
	dsp_read,
	dsp_write,
	NULL,		/* link_readdir */
	NULL,		/* link_select */
	dsp_ioctl,
	NULL,		/* link_mmap */
	dsp_open,
	dsp_release,
	NULL
};


/****************************************************************************
 *
 * This function initializes the driver. It tries to detect the hardware
 * and sets up all relevant data-structures.
 *
 ****************************************************************************/


static int DTCinitp2(void)
{
  u_int k;
  int retry,retcode;
  
	printk("DTC2 init ["__DATE__" "__TIME__"]\n");
	DSPexists = 0;
	
	if (!DTC2present()) {
		printk("DTC2 not found\n");
		retcode=1; goto quit;
	}
	FPGAerr  = inb(iXINIT)&1;
	FPGAdone = inb(iXDONE)&1;
	FPGAmode = inb(iXMODE)&1;
	printk("FPGA mode=%d err=%d done=%d - ",FPGAmode,FPGAerr,FPGAdone);

	k = inb(iXDONE)&1;
	if (k) printk("FPGA already configured\n");
	 else {
		printk("FPGA not configured\n");
		if (fpgaconfTAB()) {
			retcode=1; goto quit;
		}
	  }
	outb(8,FPGA);		/* reset MPU401 */
/*
   [0]IRQ9EN=0  [1]IRQ5EN=0  [2]EVREN=0  [3]RESET401=pulse  [4]NMIEN=0
   [5]TFIRQ=0   [6]CTRL6=1   [7]CAPTEN=0
*/
	outb(0x40,FPGA);

	use_schedule = 1;
	DSPowner = 0;
	OnCEmode = 0;
	retry = 0;
again:
	retcode = 0;
	if (QRDX->ident == 0x435444) {
		printk("DTC2 initialized");
		k = QRDX->QTIML;
		d_delay(1);
		if (k != QRDX->QTIML) {
			printk(" and running, t=%d sec",k/10000);
			DSPexists = 1;
		}
		else if ((inb(OnCEstat)&7)==4) {
			printk(" in debug mode");
			DSPexists = 1;
			OnCEmode=1;
		}
		printk("\n");
	}
	else if (QRDX->ident == 0x444e45) {
		printk("DTC2 previously stopped");
	}
	if (!DSPexists && !retry) {
		printk("...resetting...");
		k=DSPreset();
		printk("says %06x\n",k);
		retry=1;
		d_delay(0);
		goto again;
	}
	if (!DSPexists) retcode=1;

quit:
	return retcode;
}


#ifndef MODULE

long dtc2_init(long kmem_start)
{
	
	if ( register_chrdev( DSP56_MAJOR, "dsp56", &dsp_fops ) )
	{
		printk("dsp_init: unable to get major %d for 56002 interface.\n",
			DSP56_MAJOR );
		return kmem_start;
	}

	DTCinitp2();

	return kmem_start;

}

#endif

/****************************************************************************/

#ifdef MODULE

int init_module(void) {

	if (DTCinitp2()) return(-ENODEV);

	if ( register_chrdev( DSP56_MAJOR, "dsp56", &dsp_fops ) )
	{
	  printk("dsp_init: unable to get major %d for 56002 interface.\n",
		DSP56_MAJOR );
	  return -EIO;
	}

	request_region(oECF2,14,CARD_NAME);
	request_region(FPGA,4,CARD_NAME);
	request_region(M401,8,CARD_NAME);
	request_region(DSP56,12,CARD_NAME);

	return 0;
}


void cleanup_module( void) {

  if (MOD_IN_USE)
	printk("dtc2: device busy, remove delayed\n");

  if (unregister_chrdev(DSP56_MAJOR, "dsp56") != 0) {
	printk("cleanup_module failed\n");
  } else {
	QRDX->ident=0x444e45;
	outb(0,oDRES);
	outb(0,oDMODE);
	outb(1,oTRESET);
	release_region(oECF2,14);
	release_region(FPGA,4);
	release_region(M401,8);
	release_region(DSP56,12);
	printk("DTC2 cleanup_module succeeded\n");
  }    
}

#endif	/* MODULE */

/****************************************************************************/
