
#include "..\dtc2.h"
#include <bios.h>
#include <dir.h>

#define DEBUG	1
#define USEASM	1

extern int in_debug_mode;
extern int HWxmode,bootmode,Resetp,LoadFile,StopMon;
extern unsigned lowestP;

extern OnCEregs Oregs;
extern DSPregs  Dregs;

extern char far *QRAM;

static char dline[] = "-------------------------------------------------------------------";

extern ulong bk_hi,bk_lo,bk_cnt;
extern int   isbk;

/*
	01	bf	break on fetch
	02	bpf	break on P read/fetch
	03	bpa	break on P access
	04	bfe	break on executed fetches
	05	bpw	break on P write
	06	bpr	break on P read
	07	bp	break on P read/write
	09	bxw	break on X write
	0a	bxr	break on X read
	0b	bx	break on X read/write
	0d	byw	break on Y write
	0e	byr	break on Y read
	0f	by	break on Y read/write
*/
/* ----------------------------------------------------------------- */
/*
  test DTC2 present *and* FPGA initialised - no OnCE otherwise!
*/

int DTC2present(char far **QRAM)
{
 int k,seg,err;
 char *es;

  err=1;
/* reset XMODE */
  outportb(oXMODE,0);
  HWxmode = inportb(iXMODE)&1;	/* 1 if inv(PC), 0 if PROM */

/* toggle/test XMODE */
  outportb(oXMODE,1);
  if ((inportb(iXMODE)&1)^HWxmode) err=0;
  outportb(oXMODE,0);
  if ((inportb(iXMODE)&1)^HWxmode) err=0;
  if (err) return(1);		/* no DTC2, no FPGA to test */

  *QRAM=NULL;
  es=getenv("DTCQRAM");
  if (es==NULL) err=1;
  else {
    if (sscanf(es,"%x",&k)==1) {
      switch(k) {
	case 0: case 0xc8: case 0xc800: seg=0xc800; break;
	case 1: case 0xcc: case 0xcc00: seg=0xcc00; break;
	case 2: case 0xd0: case 0xd000: seg=0xd000; break;
	default: err=1;
       }
      if (!err) *QRAM=MK_FP(seg,0);
     }
   }
  if (*QRAM==NULL)
    fprintf(stderr,"Invalid DTCQRAM environment setting\n");

  return(!err);
}


int OnCEfakeboot(void)
{
 int n,t;
 ulong v;

  for (n=0; n<512; n++) {
    t=10000; while (((inp(HI_ISR)&0xf2)==0) && (--t));
    if (t==0) { fprintf(stderr,"boot:#%d\n",n); return(1); }
    switch(n) {
      case 0x00: v=0x0af080l; break;
      case 0x01: v=0x000040l; break;
      case 0x40: v=0x44f400l; break;
      case 0x41: v=0x1f000al; break;
      case 0x42: v=0x08c43dl; break;
      case 0x45: v=0x0c0043l; break;
      default:   v=0;
     }
    outp(HI_TXM,(v>>8)); outp(HI_TXM,(v>>8));
    outp(HI_TXH,(v>>16)); outp(HI_TXH,(v>>16));
    outp(HI_TXL,v);
   }
  return(0);
}


/* ================================================================= */
/*                                                                   */
/*		LOW-LEVEL OnCE interface                             */
/*                                                                   */
/* ================================================================= */
/*
  Try to guess if there was an OnCE request - look at ACK,DSCK,DSI
	if DSCK:DSI != 00 they are outputs, DSP not in debug mode
	if DSCK:DSI == 00 and ACK=0 the DSP is in normal run mode
			*OR* ACK was canceled not in debug mode (beware!)
	if DSCK:DSI == 00 and ACK=1 the DSP is in debug mode waiting
			for service *OR* someone forgot to clear ACK
			after a debug session (beware!)
*/
int OnCEreq(void)
{
  return((inportb(OnCEstat)&7)==4);
}

int OnCErunmode(void)
{
  return(inportb(OnCEstat)&3);
}


int enterOnCE(int reset)
{
 unsigned t;

  if (in_debug_mode) return(-1);
  if (reset || !OnCEreq()) {
    outportb(OnCEstat,1);		/* set DR=0, clear ACK */
    if (reset) {			/* reset processor */
      outportb(oDRES,0);		/* (confirm) DSP reset */
      delay(1);
      outportb(oDRES,1);		/* start DSP with DR=0 */
     }
    t=100000;
    while (((inportb(OnCEstat)&4)==0)&&(--t));
    outportb(OnCEstat,0);		/* set DR=1, clear ACK */
    if (t==0) return(1);
   }
  else outportb(OnCEstat,0);		/* clear request */
  in_debug_mode=1;
  return(0);
}


static int send_command(unsigned char n)
{
 unsigned t;

#ifdef USEASM
  asm {	mov	dx,349h
	mov	al,byte ptr n
	out	dx,al
	inc	dx
	cli
	out	dx,al
	sub	dx,2   }
loop0001: asm {
	in	al,dx
	test	al,8
	je	loop0001
	sti
	xor	cx,cx	}
loop0002: asm {
	in	al,dx
	db	0ebh,0
	test	al,4
	loope	loop0002  }
xit0001:  asm 	nop;
  return(_CX==0);
#else
  outportb(OnCEdat,n);		/* data latched into shift reg */
  disable();			/* no interrupts while transmitting */
/* 8 clks at PCCLK/4 -> 32 PCCLK cycles (1.92..3.84 us) */
  outportb(OnCEclk,0);		/* start 8 clock cycles */
  while ((inportb(OnCEstat)&8)==0);	/* wait for end of clock sequence */
/* command has always an ACK, look for it */
  enable();
  t=20000;
  while (((inportb(OnCEstat)&4)==0)&&(--t));
  return(t==0);
#endif
}


static int send_data24(ulong v,int ack)
{
 unsigned t;

#ifdef USEASM
  asm {	mov	dx,349h
	mov	ax,word ptr v+2
	mov	bx,word ptr v
	out	dx,al
	inc	dx
	cli
	out	dx,al
	sub	dx,2   }
loop0011: asm {
	in	al,dx
	test	al,8
	je	loop0011
	mov	al,bh
	inc	dx
	out	dx,al
	inc	dx
	out	dx,al
	sub	dx,2   }
loop0012: asm {
	in	al,dx			/* dx=stat */
	test	al,8
	je	loop0012
	mov	al,bl
	inc	dx
	out	dx,al			/* data */
	xor	ax,ax
	dec	dx			/* dx=stat */
	out	dx,al			/* clear ack */
	add	dx,2			/* dx=clock */
	out	dx,al
	sub	dx,2   }		/* dx=stat */
loop0013: asm {
	in	al,dx
	test	al,8
	je	loop0013
	sti	}
#else
/* data format: 00HHMMLL, 24 lowest bits */
  outportb(OnCEdat,(v>>16));	/* write H */
  disable();
  outportb(OnCEclk,0);		/* 8 auto cycles */
  while ((inportb(OnCEstat)&8)==0);
  outportb(OnCEdat,(v>>8));	/* write M */
  outportb(OnCEclk,0);		/* 8 auto cycles */
  while ((inportb(OnCEstat)&8)==0);
  outportb(OnCEdat,v);		/* write L */
  outportb(OnCEclk,0);		/* 8 auto cycles */
  if (!ack) outportb(OnCEstat,0);	/* keep ACK clear */
  while ((inportb(OnCEstat)&8)==0);
  enable();
#endif
  if (!ack) return(0);
  t=20000;
  while (((inportb(OnCEstat)&4)==0)&&(--t));
  return(t==0);
}


static long get_data24(int ack)
{
 unsigned t;
 ulong v;

#ifdef USEASM
  asm {	mov	dx,349h
	xor	ax,ax
	out	dx,al			/* data<-0 */
	inc	dx			/* dx=clock */
	cli
	out	dx,al
	sub	dx,2   }		/* dx=stat,ah=0 */
loop0021: asm {
	in	al,dx
	test	al,8
	je	loop0021
	inc	dx			/* dx=data,ah=0 */
	in	al,dx
	mov	word ptr v+2,ax
	mov	al,ah			/* zero al */
	out	dx,al			/* data<-0 */
	inc	dx
	out	dx,al			/* clock */
	sub	dx,2   }
loop0022: asm {
	in	al,dx
	test	al,8
	je	loop0022
	inc	dx			/* dx=data */
	in	al,dx
	mov	cl,al			/* save M byte in cl */
	mov	al,ah			/* zero al */
	out	dx,al			/* data<-0 */
	dec	dx			/* dx=stat */
	out	dx,al			/* clear ack */
	add	dx,2			/* dx=clock */
	out	dx,al
	sub	dx,2   }		/* dx=stat */
loop0023: asm {
	in	al,dx
	test	al,8
	je	loop0023
	inc	dx
	in	al,dx
	sti
	mov	ah,cl
	mov	word ptr v,ax }
#else
  outportb(OnCEdat,0);		/* keep DSI output low */
  disable();
  outportb(OnCEclk,0);		/* 8 auto cycles */
  while ((inportb(OnCEstat)&8)==0);
  v=inportb(OnCEdat);		/* get H */
  outportb(OnCEclk,0);		/* 8 auto cycles */
  while ((inportb(OnCEstat)&8)==0);
  v=(v<<8)|inportb(OnCEdat);	/* get M */
  outportb(OnCEclk,0);		/* 8 auto cycles */
  if (!ack) outportb(OnCEstat,0);	/* keep ACK clear */
  while ((inportb(OnCEstat)&8)==0);
  enable();
  v=(v<<8)|inportb(OnCEdat);	/* get L */
#endif
  if (ack) {
    t=20000;
    while (((inportb(OnCEstat)&4)==0)&&(--t)); if (t==0) return(-1l);
   }
  return(v);		/* always positive if no error */
}


/* send command and read data - do not check for ACK */

long OnCErdat(unsigned char reg)
{
  if (send_command(reg|O_read)) return(-1l);
  return(get_data24(0));
}

/* send command and write data with/without ACK */

int OnCEwdat(int reg,ulong v,int ack)
{
  if (send_command(reg|O_write)) return(1);
  return(send_data24(v,ack));
}


/* ================================================================= */
/*                                                                   */
/*                                                                   */
/* ================================================================= */


int begin_debug_run(OnCEregs *o,int reset)
{
 int i,err;

#ifdef DEBUG
  fprintf(stderr,"Enter OnCE, status=%02x\n",inportb(OnCEstat)&15);
#endif

  if (err=enterOnCE(reset)) {
    if (err<0) perror("OnCE already active");
     else perror("Timeout on DR ACK");
    return(1);
   }
#ifdef DEBUG
  fprintf(stderr,"Inside OnCE, status=%02x\n",inportb(OnCEstat)&15);
#endif

  if ((o->PDBR=OnCErdat(O_OPDBR))==-1l) return(2);
  if ((o->PILR=OnCErdat(O_OPILR))==-1l) return(3);
  if ((o->PABFR=OnCErdat(O_OPABFR))==-1l) return(4);
  if ((o->PABDR=OnCErdat(O_OPABDR))==-1l) return(5);
  for (i=0; i<5; i++)
    if ((o->PABfifo[i]=OnCErdat(O_FIFO))==-1l) return(6);
  if ((o->SCR=OnCErdat(O_OSCR))==-1l) return(7);

  return(0);
}


int end_debug_go(OnCEregs *o,ulong a)	/* if a==0 no jump */
{
  OnCEwdat(O_OSCR,0,0);		/* no trace mode */
  if (WritePDBAck(a? O_jmplong:o->PILR)) return(10);
  if (WritePDBgoex(a? a:o->PDBR)) return(11);	/* address right aligned (M,L) */
  in_debug_mode=0;

#ifdef DEBUG
  fprintf(stderr,"Exit OnCE");
  if (a) fprintf(stderr," and jump %04lx",a);
  fprintf(stderr,", status=%02x\n",inportb(OnCEstat)&15);
#endif
/*  outportb(OnCEstat,0);		clear ACK */
  return(0);
}


int OnCEtrace(OnCEregs *o,ulong n,ulong a)
{
 int i,stat;
 ulong t,v;

  OnCEwdat(O_OTC,n-1,0);
  OnCEwdat16(O_OSCR,o->SCR|O_OSCR_TME,0);
  v=OnCErdat(O_OSCR);
  if (WritePDBAck(a==0xffff? o->PILR:O_jmplong)) return(10);
  if (WritePDBgoex(a==0xffff? o->PDBR:a)) return(11);

  t=200000;
  while (((stat=(inportb(OnCEstat)&7))!=4)&&(--t)) {
/*    if (stat&3) { in_debug_mode=0; enterOnCE(0); } */
   }
  if (t==0) fprintf(stderr,"Trace stat=%x\n",stat);

  if ((o->SCR=OnCErdat(O_OSCR))==-1l) return(7);
/*  if ((o->SCR&0x40000)==0) return(8); */
  if ((o->PDBR=OnCErdat(O_OPDBR))==-1l) return(2);
  if ((o->PILR=OnCErdat(O_OPILR))==-1l) return(3);
  if ((o->PABFR=OnCErdat(O_OPABFR))==-1l) return(4);
  if ((o->PABDR=OnCErdat(O_OPABDR))==-1l) return(5);
  for (i=0; i<5; i++)
    if ((o->PABfifo[i]=OnCErdat(O_FIFO))==-1l) return(6);

  return(0);
}


int OnCEbrkgo(OnCEregs *o)
{
 volatile int i,stat;
 ulong v;

  v=OnCErdat(O_OSCR);
  OnCEwdat(O_OMBC,bk_cnt-1,0);
  OnCEwdat(O_OMLLR,bk_lo,0);
  OnCEwdat(O_OMULR,bk_hi,0);
  OnCEwdat16(O_OSCR,isbk,0);
  if (WritePDBAck(o->PILR)) return(10);
  if (WritePDBgoex(o->PDBR)) return(11);

  while ((stat=(inportb(OnCEstat)&7))!=4) {
    if (i=bioskey(1)) { i=bioskey(0); return(12); }
   }

  if ((o->SCR=OnCErdat(O_OSCR))==-1l) return(7);
/*  if ((o->SCR&0x30000)==0) return(8); */
  OnCEwdat(O_OSCR,v,0);
  if ((o->PDBR=OnCErdat(O_OPDBR))==-1l) return(2);
  if ((o->PILR=OnCErdat(O_OPILR))==-1l) return(3);
  if ((o->PABFR=OnCErdat(O_OPABFR))==-1l) return(4);
  if ((o->PABDR=OnCErdat(O_OPABDR))==-1l) return(5);
  for (i=0; i<5; i++)
    if ((o->PABfifo[i]=OnCErdat(O_FIFO))==-1l) return(6);

  return(0);
}


int OnCEdebgo(OnCEregs *o)
{
 volatile int i,stat;
 ulong v;

  OnCEwdat(O_OSCR,0,0);		/* no trace mode */
  if (WritePDBAck(o->PILR)) return(10);
  if (WritePDBgoex(o->PDBR)) return(11);

  while ((stat=(inportb(OnCEstat)&7))!=4) {
    if (i=bioskey(1)) { i=bioskey(0); return(12); }
   }

  if ((o->SCR=OnCErdat(O_OSCR))==-1l) return(7);
  if ((o->PDBR=OnCErdat(O_OPDBR))==-1l) return(2);
  if ((o->PILR=OnCErdat(O_OPILR))==-1l) return(3);
  if ((o->PABFR=OnCErdat(O_OPABFR))==-1l) return(4);
  if ((o->PABDR=OnCErdat(O_OPABDR))==-1l) return(5);
  for (i=0; i<5; i++)
    if ((o->PABfifo[i]=OnCErdat(O_FIFO))==-1l) return(6);

  return(0);
}


/* ------------------------------------------------------------------- */


/* User registers */
ulong OnCErreg(int r)		/* r = register NUM */
{
  Execute1i(O_GmovR|(r<<8));
  return(ReadGDB());
}

int OnCEwreg(int r,ulong v)	/* r = register NUM */
{
 ulong opr;

  opr=r;
  if (r<reg_M0) {
    if (r>=reg_N0) opr+=0x18;
     else if (r>=reg_R0) opr+=0x10;
      else if (r>=reg_A0) opr+=8;
    Execute2i1(O_ImovR|(opr<<16));
   }
  else {
    Execute2i1(O_ImovM|r);
   }
  return(Execute2i2(v));
}


int OnCEgetallreg(DSPregs *r,ulong mask)
{
 int i;

  if (mask&msk_X0) r->X0=OnCErreg(reg_X0);
  if (mask&msk_X1) r->X1=OnCErreg(reg_X1);
  if (mask&msk_Y0) r->Y0=OnCErreg(reg_Y0);
  if (mask&msk_Y1) r->Y1=OnCErreg(reg_Y1);
  if (mask&msk_A2) r->A2=OnCErreg(reg_A2);
  if (mask&msk_A1) r->A1=OnCErreg(reg_A1);
  if (mask&msk_A0) r->A0=OnCErreg(reg_A0);
  if (mask&msk_B2) r->B2=OnCErreg(reg_B2);
  if (mask&msk_B1) r->B1=OnCErreg(reg_B1);
  if (mask&msk_B0) r->B0=OnCErreg(reg_B0);
  for (i=0; i<8; i++) {
    if (mask&msk_RNM(i)) {
      r->R[i]=OnCErreg(i+0x10);
      r->N[i]=OnCErreg(i+0x18);
      r->M[i]=OnCErreg(i+0x20);
     }
   }
  if (mask&msk_SR) r->SR=OnCErreg(reg_SR);
  if (mask&msk_OMR) r->OMR=OnCErreg(reg_OMR);
  if (mask&msk_LA) r->LA=OnCErreg(reg_LA);
  if (mask&msk_LC) r->LC=OnCErreg(reg_LC);
  if (mask&msk_SP) r->SP=OnCErreg(reg_SP);
  return(0);
}


int OnCErmem(char space,ulong a,int n,ulong *buf)
{
 ulong r0;		/* save R0 value */
 ulong Smove,Mask24;

  if (n==0) return(0);
  Mask24=0xffffffl;
  switch(space) {
    case 'x': Smove = O_movR0p+0x40; break;
    case 'y': Smove = O_movR0p+0x80; break;
    case 'p': Smove = O_movR0p+0x00; break;
    default:  return(1);
   }
  r0=OnCErreg(reg_R0);
  Execute2i1(O_moviR0);			/* move #aaaa,r0 */
  Execute2i2(a);
  Execute1i(Smove);			/* move S:(r0)+,X:OGDB */
  *buf++ = ReadGDB()&Mask24; n--;
  while (n--) {
    NOPgo(); *buf++ = ReadGDB()&Mask24;
   }
  OnCEwreg(reg_R0,r0);
  return(0);
}



int OnCEwmem(char space,ulong a,int n,ulong *buf)
{
 ulong r0,x0;		/* save R0,X0 value */
 ulong Smove;

  if (n==0) return(0);
  switch(space) {
    case 'x': Smove = 0x445800l; break;		/* move x0,X:(r0)+ */
    case 'y': Smove = 0x4c5800l; break;		/* move x0,Y:(r0)+ */
    case 'p': Smove = 0x075884l; break;		/* move x0,P:(r0)+ */
    default:  return(1);
   }
  r0=OnCErreg(reg_R0);
  x0=OnCErreg(reg_X0);
  Execute2i1(O_moviR0);			/* move #aaaa,r0 */
  Execute2i2(a);
  Execute2i1(O_moviX0);			/* move #nnnn,x0 */
  Execute2i2(*buf++);
  Execute1i(Smove); n--;		/* move x0,d:(r0)+ */
  while (n--) {
    Execute2i1(O_moviX0);		/* move #nnnn,x0 */
    Execute2i2(*buf++);
    Execute1i(Smove);	 		/* move x0,d:(r0)+ */
   }
  OnCEwreg(reg_X0,x0);
  OnCEwreg(reg_R0,r0);
  return(0);
}


int OnCEfillmem(char space,ulong a,int n,ulong v)
{
 ulong r0,x0;		/* save R0,X0 value */
 ulong Smove;

  if (n==0) return(0);
  switch(space) {
    case 'x': Smove = 0x445800l; break;		/* move x0,X:(r0)+ */
    case 'y': Smove = 0x4c5800l; break;		/* move x0,Y:(r0)+ */
    case 'p': Smove = 0x075884l; break;		/* move x0,P:(r0)+ */
    default:  return(1);
   }
  r0=OnCErreg(reg_R0);
  x0=OnCErreg(reg_X0);
  Execute2i1(O_moviR0);			/* move #aaaa,r0 */
  Execute2i2(a);
  Execute2i1(O_moviX0);			/* move #nnnn,x0 */
  Execute2i2(v);
  Execute1i(Smove); n--;		/* move x0,d:(r0)+ */
  while (n--) {
    NOPgo();
   }
  OnCEwreg(reg_X0,x0);
  OnCEwreg(reg_R0,r0);
  return(0);
}


/* ================================================================= */
/*                                                                   */
/*                                                                   */
/* ================================================================= */


static rectyp findkeyw (char *t)
{
 char typessym[6][12] =
      { "START","END","DATA","BLOCKDATA","SYMBOL","COMMENT" };
 int i,found;

  found=0;
  for (i=0; (i<6)&&(!found); i++) {
    found = !stricmp(t,typessym[i]);
   }
  if (!found) return(sNorec);
   else return((rectyp)(i-1));
}


unsigned OnCEloadDSP(char *fname,ulong *dspbuf)
{
 char drive[MAXDRIVE];
 char dir[MAXDIR];
 char file[MAXFILE];
 char ext[MAXEXT];
 int flags;
 char t[132];
 int i;
 rectyp blktype = sNorec;
 char mspace[3],s;
 unsigned len,err;
 ulong *buf;
 ulong start,V;
 FILE *fp;

  flags=fnsplit(fname,drive,dir,file,ext);
  fnmerge(t,drive,dir,file,".lod");
  if ((fp = fopen(t,"rt")) == NULL) {
    fprintf(stderr,"No file %s\n",t);
    return(1);
   }
  lowestP=0xffff;

  do {
   if (fscanf(fp,"%s",t)==1) {
     if (*t == '_') {
       blktype = findkeyw(&(t[1]));
       switch (blktype) {
	 case sData:
	     fscanf(fp,"%s",mspace); s=tolower(*mspace);
	     fscanf(fp,"%4lX",&start);
	     len=0; i=1; buf=dspbuf;
	     while (i) {
	       i = fscanf(fp,"%6lX",buf);
	       if (i) { buf++; len++; }
	      }
#ifdef DEBUG
	     fprintf(stderr,"\tblk %c:%04lx l=%x\n",s,start,len);
#endif
	     OnCEwmem(s,start,len,dspbuf);
	     if (s=='p') { if (start<lowestP) lowestP=start; }
	     break;
	 case sBlock:
	     fscanf(fp,"%s",mspace); s=tolower(*mspace);
	     fscanf(fp,"%4lX",&start);
	     fscanf(fp,"%4x",&len);
	     fscanf(fp,"%6lX",&V);
#ifdef DEBUG
	     fprintf(stderr,"\tfill %06lx %c:%04lx l=%x\n",V,s,start,len);
#endif
	     OnCEfillmem(s,start,len,V);
	     break;
	 default: break;
	}
      }
    }
  } while (!(feof(fp) || (blktype==sEnd)));
 fclose(fp);
#ifdef DEBUG
 fprintf(stderr,"Loaded %s at P:%04x\n",fname,lowestP);
#endif

 return(lowestP);
}


/* ------------------------------------------------------------------- */

static ulong kPIL(ushort a,int n)
{
  return(n? Oregs.PDBR:Oregs.PILR);
}


void OnCEshowallreg(DSPregs *r,OnCEregs *o,FILE *fe)
{
 int i,k;
 ushort a;
 ulong v,s;

  fprintf(fe,"%s\n",dline);
  v=OnCErdat(O_OTC);
  s=o->SCR>>8;
  fprintf(fe,"SCR: %04lx  Trace=%d  Brk mode=%lx  TC=%06lx\n",s,
	((s&O_OSCR_TME)!=0),s&O_OSCR_BC,v);
  fprintf(fe,"FIFO:");
  for (i=0; i<5; i++) fprintf(fe," %04lx",(o->PABfifo[i]>>8));
  fprintf(fe,"  Fetch=%04lx  PDB=%06lx\n",(o->PABFR>>8),o->PDBR);
  fprintf(fe,"   X1:X0=   %06lx:%06lx\t",r->X1,r->X0);
  fprintf(fe,"   Y1:Y0=   %06lx:%06lx\n",r->Y1,r->Y0);
  fprintf(fe,"A2:A1:A0=%02x:%06lx:%06lx\t",r->A2,r->A1,r->A0);
  fprintf(fe,"B2:B1:B0=%02x:%06lx:%06lx\n",r->B2,r->B1,r->B0);
  for (i=0; i<4; i++) {
    k=i;
    fprintf(fe,"   R%d=%04x  N%d=%04x  M%d=%04x\t",k,r->R[k],
	k,r->N[k],k,r->M[k]);
    k=i+4;
    fprintf(fe,"   R%d=%04x  N%d=%04x  M%d=%04x\n",k,r->R[k],
	k,r->N[k],k,r->M[k]);
   }
  fprintf(fe,"SR=%04x  OMR=%04x  ",r->SR,r->OMR);
  fprintf(fe,"SP=%04x   LA=%04x  ",r->SP,r->LA);
  fprintf(fe,"LC=%04x\n",r->LC);
  fprintf(fe,"%s\n",dline);

  a=o->PABDR>>8;
  disasm56(a,kPIL,fe);
}


/* ----------------------------------------------------------------- */

static char regnam1[] =
	"x0 x1 y0 y1 a0 b0 a2 b2 a1 b1 a  b  r0 r1 r2 r3 r4 r5 r6 r7 "
	"n0 n1 n2 n3 n4 n5 n6 n7 m0 m1 m2 m3 m4 m5 m6 m7 ";
static char regnam2[] =
	"sr  omr sp  ssh ssl la  lc  ";

int OnCEgetregn(char *p,FILE *fe)
{
 char *s;
 int n;
 ulong v;

  strlwr(p);
  if (strlen(p)==1) {
    switch(*p) {
      case 'a': fprintf(fe,"A = %02lx:",OnCErreg(reg_A2)&0xff);
		fprintf(fe,"%06lx:%06lx\n",OnCErreg(reg_A1),OnCErreg(reg_A0));
		break;
      case 'b': fprintf(fe,"B = %02lx:",OnCErreg(reg_B2)&0xff);
		fprintf(fe,"%06lx:%06lx\n",OnCErreg(reg_B1),OnCErreg(reg_B0));
		break;
      case 'x': fprintf(fe,"X = %06lx:%06lx\n",OnCErreg(reg_X1),OnCErreg(reg_X0));
		break;
      case 'y': fprintf(fe,"Y = %06lx:%06lx\n",OnCErreg(reg_Y1),OnCErreg(reg_Y0));
		break;
      default : fprintf(stderr,"Bad register: %s\n",p); return(-1);
     }
    return(0);
   }

  if ((s=strstr(regnam1,p))!=NULL) {
    n = ((s-regnam1)/3)+4;
   }
  else if ((s=strstr(regnam2,p))!=NULL) {
    n = ((s-regnam2)/4)+0x39;
   }
  else {
    fprintf(stderr,"Bad register: %s\n",p); return(-1);
   }

  v = OnCErreg(n);
  if (n>=reg_R0)
    printf("%c%c%c=%04lx\n",s[0],s[1],s[2],(v&0xffff));
  else
    printf("%c%c%c=%06lx\n",s[0],s[1],s[2],v);
  return(0);
}


/* ----------------------------------------------------------------- */


int OnCEputregn(char *p,ulong v)
{
 char *s;
 int n;
 ulong vc;

  strlwr(p);
  if (strlen(p)==1) {
    switch(*p) {
      case 'a': vc=(v&0x800000l? 0xff:0);
		OnCEwreg(reg_A2,vc); OnCEwreg(reg_A1,v);
		OnCEwreg(reg_A0,0);
		return(0);
      case 'b': vc=(v&0x800000l? 0xff:0);
		OnCEwreg(reg_B2,vc); OnCEwreg(reg_B1,v);
		OnCEwreg(reg_B0,0);
		return(0);
      case 'x':
      case 'y': fprintf(stderr,"Can change only X/Y0,X/Y1\n"); return(0);
      default : fprintf(stderr,"Bad register: %s\n",p); return(-1);
     }
    return(0);
   }

  if ((s=strstr(regnam1,p))!=NULL) {
    n = ((s-regnam1)/3)+4;
   }
  else if ((s=strstr(regnam2,p))!=NULL) {
    n = ((s-regnam2)/4)+0x39;
   }
  else {
    fprintf(stderr,"Bad register: %s\n",p); return(-1);
   }

  OnCEwreg(n,v);
  vc = OnCErreg(n);
  if (v!=vc) {
    fprintf(stderr,"Register not correctly set: %06lx\n",vc);
    return(-1);
   }
  return(0);
}


/* ----------------------------------------------------------------- */

