
#include "dsp_type.h"

/* -------------------------------------------------------------------- */

#define OnCEdat		(OnCE_BASE)		/* clocked */
#define OnCEndat	(OnCE_BASE+4)		/* not clocked */
#define OnCEack		(OnCE_BASE+5)

/* OnCE command register - bits 0-4=register no. */
#define O_read		0x80
#define O_write		0x00
#define O_go		0x40
#define O_nogo		0x00
#define O_ex		0x20
#define O_noex		0x00

/* OnCE registers */
#define O_OSCR		0x00	/* status/control register */
#define O_OMBC		0x01	/* memory breakpoint counter */
#define O_OTC		0x03	/* trace counter */
#define O_OMULR		0x06	/* memory upper limit register */
#define O_OMLLR		0x07	/* memory lower limit register */
#define O_OGDBR		0x08	/* global data bus register */
#define O_OPDBR		0x09	/* program data bus register */
#define O_OPABFR	0x0A	/* PAB register for fetch */
#define O_OPILR		0x0B	/* pipeline instruction latch register */
#define O_OCMBC		0x0C	/* clear memory breakpoint counter */
#define O_OCTC		0x0E	/* clear trace counter */
#define O_FIFO		0x11
#define O_OPABDR	0x13	/* PAB register for decode */
#define O_noreg		0x1F

/* OnCE status/control register 0x00 */
#define O_OSCR_TO	0x400	/* trace occurrence, R/O */
#define O_OSCR_MBO	0x200	/* memory breakpoint occurrence, R/O */
#define O_OSCR_SWO	0x100	/* SW debug occurrence, R/O */
#define O_OSCR_TME	0x10	/* Trace mode enable, R/W */
#define O_OSCR_BC	0x0f	/* mask for breakpoint mode */

/* breakpoint modes */
#define O_nobrk		0x0
#define O_brk_fetch	0x1	/* fetch+aborted instructions */
#define O_brk_readPf	0x2	/* read P+fetch */
#define O_brk_accessP	0x3	/* read+write+fetch P */
#define O_brk_xfetch	0x4	/* executed fetches */
#define O_brk_writeP	0x5	/* write P */
#define O_brk_readP	0x6	/* read P, no fetch */
#define O_brk_rwP	0x7	/* read+write P */
#define O_brk_writeX	0x9
#define O_brk_readX	0xa
#define O_brk_rwX	0xb
#define O_brk_writeY	0xd
#define O_brk_readY	0xe
#define O_brk_rwY	0xf


#define O_GmovR		0x08C03Cl	/* movep reg,X:OGDB */
/* OR with register - order:
	0x400	x0,x1,y0,y1
	0x800	a0,b0,a2,b2,a1,b1,a,b
	0x1000	r0..r7
	0x1800	n0..n7
	0x2000	m0..m7
	0x2800	-------
	0x3900	sr,omr,sp,ssh,ssl,la,lc
*/
#define O_reg_X0	0x04		/* 0x00000002 */
#define O_reg_X1	0x05		/* 0x00000001 */
#define O_reg_Y0	0x06		/* 0x00000008 */
#define O_reg_Y1	0x07		/* 0x00000004 */
#define O_reg_A0	0x08		/* 0x00000040 */
#define O_reg_B0	0x09		/* 0x00000200 */
#define O_reg_A2	0x0a		/* 0x00000010 */
#define O_reg_B2	0x0b		/* 0x00000080 */
#define O_reg_A1	0x0c		/* 0x00000020 */
#define O_reg_B1	0x0d		/* 0x00000100 */
#define O_reg_A 	0x0e
#define O_reg_B 	0x0f
#define O_reg_R(r)	(0x10+(r))	/* 0x0003FC00 */
#define O_reg_N(r)	(0x18+(r))	/* 0x0003FC00 */
#define O_reg_M(r)	(0x20+(r))	/* 0x0003FC00 */
#define O_reg_SR	0x39		/* 0x00040000 */
#define O_reg_OMR	0x3a		/* 0x00080000 */
#define O_reg_SP	0x3b		/* 0x00100000 */
#define O_reg_SSH	0x3c		/* 0x00200000: do not read! */
#define O_reg_SSL	0x3d		/* 0x00400000: do not read! */
#define O_reg_LA	0x3e		/* 0x00800000 */
#define O_reg_LC	0x3f		/* 0x01000000 */

#define O_moviR0	0x60F400l	/* move #xxxx,r0   */
#define O_moviX0	0x44F400l	/* move #xxxxxx,x0 */

#define O_movR0p	0x08D87Cl	/* move S:(r0)+,X:OGDB */
#define O_movAp		0x08F07Cl	/* move abs,X:OGDB     */
/* OR with  P=0x00  X=0x40  Y=0x80 */

/* -------------------------------------------------------------------- */

