
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <dir.h>
#include <dos.h>

#define oTRESET		0x160
#define oTANALY		0x161
#define oECF2		0x162
#define oECF3		0x163
#define oXMODE		0x164
#define oDRES		0x165
#define oDMODE		0x166
#define oXPROG		0x167

#define FPGA		0x178
#define DSP56		0x340
#define ONCE		0x348
#define C012		0x150

#define IDECS0		0x170
#define IDECS1		0x374

#define iXINIT		0x160
#define iXDONE		0x164
#define iXMODE		0x168
#define iLOCAL		0x16c

/* ----------------------------------------------------------------- */

#define HI_ICR		(DSP56)
#define HI_CVR		(DSP56+1)
#define HI_ISR		(DSP56+2)
#define HI_IVR		(DSP56+3)
#define HI_TXH 		(DSP56+5)
#define HI_TXM 		(DSP56+6)
#define HI_TXL 		(DSP56+7)
#define HI_RXH 		HI_TXH
#define HI_RXM          HI_TXM
#define HI_RXL          HI_TXL

#define MS_X	0
#define MS_Y	1
#define MS_P	2
#define MS_R	3
#define MS_E	4

#define HC_Peek		0x12
#define HC_Poke		0x13
#define HC_Fill		0x14

typedef enum { sStart,sEnd,sData,sBlock,sSymb,sComm,sNorec } rectyp;

char filen[128];
FILE *fs;

unsigned long *DSPBUF;
unsigned int HasP, AdrP, runP;

/* ----------------------------------------------------------------- */

#include "stub.i"

/* ----------------------------------------------------------------- */

static long near dspwrite(unsigned long v)
{
 long t;

  t=10000;
  while (((inp(HI_ISR)&2)==0)&&(t--)); if (t==0) return(-1);
  asm {	mov	dx,345h;
	mov	ax,word ptr v+2;
	out	dx,al;
	mov	ax,word ptr v;
	xchg	al,ah;
	inc	dx;
	out	dx,al;
	xchg	al,ah;
	inc	dx;
	out	dx,al;
      }
  return(0);
}


static long near dspread(void)
{
 long v,t;

  t=10000;
  while (((inp(HI_ISR)&1)==0)&&(t--)); if (t==0) return(-1);
  asm {	mov	dx,345h;
	in	al,dx;
	xor	cx,cx;
	mov	cl,al;
	inc	dx;
	in	al,dx;
	mov	ah,al;
	inc	dx;
	in	al,dx;
	mov	word ptr v,ax;
	mov	word ptr v+2,cx;
       }
  return(v);
}

/* ----------------------------------------------------------------- */
/*	LOD file read						     */
/* ------------------------------------------------------------------- */

/* stub protocol */

static int xmit56(char mspace,long start,unsigned int len)
{
 unsigned long cks,ctst,V,mfs;
 unsigned long *buf=DSPBUF;

  switch(toupper(mspace)) {
    case 'P':	if (!HasP) { HasP=1; AdrP=start; }
    case 'X':	mfs=0; goto xml2;
    case 'Y':	mfs=0x10000l;
xml2:		if (dspwrite((start&0xffffl)|mfs)) return(1);
		if (dspwrite(len)) return(2);
		cks=0;
		while (len--) {
		  V=*buf++; cks+=V;
		  if (dspwrite(V)) return(3);
		 }
		ctst=dspread();
		if (ctst==-1l) return(4);
		if ((ctst^cks)&0xffffffl) return(5);
    default:	break;
   }
  return(0);
}


static rectyp findkeyw (char *t)
{
 char typessym[6][12] =
      { "START","END","DATA","BLOCKDATA","SYMBOL","COMMENT" };
 int i,found;

  found=0;
  for (i=0; (i<6)&&(!found); i++) {
    found = !stricmp(t,typessym[i]);
   }
  if (!found) return(sNorec);
   else return((rectyp)(i-1));
}


int loadDSP(char *fname)
{
 char t[81];
 int i;
 rectyp blktype = sNorec;
 char mspace[3];
 unsigned int len,err;
 unsigned long *buf;
 long start,V;
 FILE *fp;

  if ((fp = fopen(fname,"rt")) == NULL) return(1);
  HasP=0;

  do {
   if (fscanf(fp,"%s",t)==1) {
     if (*t == '_') {
       blktype = findkeyw(&(t[1]));
       switch (blktype) {
	 case sData:
	     fscanf(fp,"%s",mspace);
	     fscanf(fp,"%4lX",&start);
	     len=0; i=1; buf=DSPBUF;
	     while (i) {
	       i = fscanf(fp,"%6lX",buf);
	       if (i) { buf++; len++; }
	      }
	     if (err=xmit56(*mspace,start,len)) return(err);
	     break;
	 case sBlock:
	     fscanf(fp,"%s",mspace);
	     fscanf(fp,"%4lX",&start);
	     fscanf(fp,"%4x",&len);
	     fscanf(fp,"%6lX",&V);
	     buf=DSPBUF;
	     for (i=0; i<len; i++) *buf++=V;
	     if (err=xmit56(*mspace,start,len)) return(err);
	     break;
	 default: break;
	}
      }
    }
  } while (!(feof(fp) || (blktype==sEnd)));
 fclose(fp);
 return(0);
}


static int bootDSP(char *stub, int l)
{
 int n,t;
 unsigned char *p;

  outportb(oDMODE,0);
  outportb(oECF2,0);
  outportb(oECF3,0);
  outportb(oDRES,0);	/* reset DSP  */
  delay(100);
  outportb(oDRES,1);	/* start DSP  */

  p = stub;
  for (n=0; (n<512)&&(l>0); n++) {
    t=10000; while (((inp(HI_ISR)&2)==0) && (t--));
    if (t==0) { printf("#%d\n",n); return(1); }
    outp(HI_TXH,*p++); outp(HI_TXM,*p++); outp(HI_TXL,*p++);
    l-=3;
   }
  for (; n<512; n++) {
    t=10000; while (((inp(HI_ISR)&2)==0) && (t--));
    if (t==0) { printf("#%d\n",n); return(1); }
    outp(HI_TXH,0); outp(HI_TXM,0); outp(HI_TXL,0);
   }
  return(0);
}


/* ------------------------------------------------------------------- */


void main (int argc, char *argv[])
{
 int err;
 char drive[MAXDRIVE];
 char dir[MAXDIR];
 char file[MAXFILE];
 char ext[MAXEXT];
 int flags;
 char *p;

  DSPBUF = (unsigned long *)calloc(1,60000);
  if (DSPBUF==NULL) {
    printf("Malloc error\n"); exit(1);
   }
  *filen=0;
  if (argc<2) {
    printf("Use: DSPLOAD filename[.lod] <-run>\n"); exit(1);
   }
  flags=fnsplit(argv[1],drive,dir,file,ext);
  fnmerge(filen,drive,dir,file,".lod");
  if ((fs=fopen(filen,"rb"))==NULL) {
    printf("File not found %s\n",filen); exit(1);
   }
  fclose(fs);
  runP=0;
  if (argc>2) {
    p=argv[2];
    runP = ((*p=='-')&&(tolower(p[1])=='r'));
   }

  err=1;
/* XMODE reset to 1 */
  outportb(oXMODE,0);	/* inv */
  if (inportb(iXMODE)&1) {
/* toggle/test XMODE */
    outportb(oXMODE,1);	/* inv */
    if (!(inportb(iXMODE)&1)) err=0;
    outportb(oXMODE,0);	/* inv */
   }
  if (err) {
    printf("DTC2 not found\n");
    goto tst1;
    exit(1);
   }

  if (bootDSP(DSPTAB,sizeof(DSPTAB))) {
    printf("Error in DSP preboot\n");
    exit(1);
   }
  else printf("DSP preboot OK - loading %s\n",filen);

tst1:
  err=loadDSP(filen);
  switch (err) {
    case 1:  printf("Start  not transmitted\n"); break;
    case 2:  printf("Length not transmitted\n"); break;
    case 3:  printf("Value  not transmitted\n"); break;
    case 4:  printf("Cksum  not received\n"); break;
    case 5:  printf("Bad checksum\n"); break;
   }
  if (err) {
    printf("Load aborted, resetting DSP\n");
    outportb(oDRES,0);	/* reset DSP  */
    exit(1);
   }
  if (HasP && (AdrP<0x100)) {
    printf("Address P:%04x too low, cannot run\n",AdrP);
    HasP = 0;
   }
  if (runP && HasP) {
    dspwrite(0x40000|HasP);
    printf("Load terminated, running from P:%04x\n",HasP);
   }
  else {
    dspwrite(0x80000);
    printf("Load terminated, not run\n");
   }

  exit(0);
}

